from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import (Any, Dict, List, Optional, Tuple, TypeAlias, TypedDict,
                    Union)


class Interval(Enum):
    ONE_MINUTE = "1min"
    FIVE_MINUTES = "5min"
    FIFTEEN_MINUTES = "15min"
    ONE_HOUR = "1h"
    FOUR_HOURS = "4h"
    ONE_DAY = "1d"
    ONE_WEEK = "1w"

Nonce: TypeAlias = int
OrderId: TypeAlias = int

class HibachiApiError(Exception):
    status_code: int
    message: str

    def __init__(self, status_code: int, message: str):
        self.status_code = status_code
        self.message = message

class TWAPQuantityMode(Enum):
    FIXED = "FIXED"
    RANDOM = "RANDOM"

class TWAPConfig:
    duration_minutes: int
    quantity_mode: TWAPQuantityMode

    def __init__(self, duration_minutes: int, quantity_mode: TWAPQuantityMode):
        self.duration_minutes = duration_minutes
        self.quantity_mode = quantity_mode

    def to_dict(self) -> Dict[str, Any]:
        return {
            "twapDurationMinutes": self.duration_minutes,
            "twapQuantityMode": self.quantity_mode.value
        }

class OrderType(Enum):
    LIMIT = "LIMIT"
    MARKET = "MARKET"
    # SCHEDULED_TWAP = "SCHEDULED_TWAP"

class Side(Enum):
    BID = "BID"
    ASK = "ASK"
    SELL = "SELL"
    BUY = "BUY"

class OrderStatus(Enum):
    PENDING = "PENDING"
    FILLED = "FILLED"
    CANCELLED = "CANCELLED"
    REJECTED = "REJECTED"
    SCHEDULED_TWAP = "SCHEDULED_TWAP"
    PLACED = "PLACED"
    PARTIALLY_FILLED = "PARTIALLY_FILLED"

class OrderFlags(Enum):
    PostOnly = "POST_ONLY"
    Ioc = "IOC"
    ReduceOnly = "REDUCE_ONLY"

@dataclass
class Order:
    accountId: int
    availableQuantity: str
    contractId: Optional[int]
    creationTime: Optional[int]
    finishTime: Optional[int]
    numOrdersRemaining: Optional[int]
    numOrdersTotal: Optional[int]
    orderFlags: Optional[OrderFlags]
    orderId: str
    orderType: OrderType
    price: Optional[str]
    quantityMode: Optional[str]
    side: Side
    status: OrderStatus
    symbol: str
    totalQuantity: Optional[str]
    triggerPrice: Optional[str]

    def __init__(self, 
                 accountId: int, 
                 availableQuantity: str, 
                 orderId: str,
                 orderType: str, 
                 side: str, 
                 status: str,
                 symbol: str,  
                 numOrdersRemaining: Optional[int] = None,
                 numOrdersTotal: Optional[int] = None, 
                 quantityMode: Optional[str] = None,
                 finishTime: Optional[int] = None,
                 price: Optional[str] = None, 
                 totalQuantity: Optional[str] = None,
                 creationTime: Optional[int] = None,
                 contractId: Optional[int] = None,
                 orderFlags: Optional[str] = None,
                 triggerPrice: Optional[str] = None):
        self.accountId = accountId
        self.availableQuantity = availableQuantity
        self.contractId = contractId
        self.creationTime = creationTime
        self.finishTime = finishTime
        self.numOrdersRemaining = numOrdersRemaining
        self.numOrdersTotal = numOrdersTotal
        self.orderId = orderId
        self.orderType = OrderType(orderType)
        self.price = price
        self.quantityMode = quantityMode
        self.side = Side(side)
        self.status = OrderStatus(status)
        self.symbol = symbol
        self.totalQuantity = totalQuantity
        self.triggerPrice = triggerPrice
        self.orderFlags = OrderFlags(orderFlags) if orderFlags else None



@dataclass
class FeeConfig:
    depositFees: str
    instantWithdrawDstPublicKey: str
    instantWithdrawalFees: List[List[Union[int, float]]]
    tradeMakerFeeRate: str
    tradeTakerFeeRate: str
    transferFeeRate: str
    withdrawalFees: str

@dataclass
class FutureContract:
    displayName: str
    id: int
    maintenanceFactorForPositions: str
    marketCloseTimestamp: Optional[str]
    marketOpenTimestamp: Optional[str]
    minNotional: str
    minOrderSize: str
    orderbookGranularities: List[str]
    riskFactorForOrders: str
    riskFactorForPositions: str
    settlementDecimals: int
    settlementSymbol: str
    status: str
    stepSize: str
    symbol: str
    tickSize: str
    underlyingDecimals: int
    underlyingSymbol: str

@dataclass
class WithdrawalLimit:
    lowerLimit: str
    upperLimit: str

@dataclass
class MaintenanceWindow:
    begin: float
    end: float
    note: str

@dataclass
class ExchangeInfo:
    feeConfig: FeeConfig
    futureContracts: List[FutureContract]
    instantWithdrawalLimit: WithdrawalLimit
    maintenanceWindow: List[MaintenanceWindow]
    # can be NORMAL, MAINTENANCE
    status: str

@dataclass
class FundingRateEstimation:
    estimatedFundingRate: str
    nextFundingTimestamp: int


@dataclass
class PriceResponse:
    askPrice: str
    bidPrice: str
    fundingRateEstimation: FundingRateEstimation
    markPrice: str
    spotPrice: str
    symbol: str
    tradePrice: str


@dataclass
class BatchResponseOrder:
    nonce: Optional[Nonce]
    orderId: Optional[OrderId]
    
    def __init__(self, 
                 nonce: Optional[Nonce] = None, 
                 orderId: Optional[OrderId] = None):
        self.nonce = int(nonce) if isinstance(nonce, str) else nonce
        self.orderId = int(orderId) if isinstance(orderId, str) else orderId


@dataclass
class BatchResponse:
    orders: List[BatchResponseOrder]

@dataclass
class StatsResponse:
    high24h: str
    low24h: str
    symbol: str
    volume24h: str

class TakerSide(Enum):
    Buy = "Buy"
    Sell = "Sell"

@dataclass
class Trade:
    price: str
    quantity: str
    takerSide: TakerSide
    timestamp: int

@dataclass
class TradesResponse:
    trades: List[Trade]

@dataclass
class Kline:
    close: str
    high: str
    low: str
    open: str
    interval: str
    timestamp: int
    volumeNotional: str

@dataclass
class KlinesResponse:
    klines: List[Kline]

@dataclass
class OpenInterestResponse:
    totalQuantity: str

@dataclass
class OrderBookLevel:
    price: str
    quantity: str

@dataclass
class OrderBook:
    ask: List[OrderBookLevel]
    bid: List[OrderBookLevel]

@dataclass
class Asset:
    quantity: str
    symbol: str

@dataclass
class Position:
    direction: str
    entryNotional: str
    markPrice: str
    notionalValue: str
    openPrice: str
    quantity: str
    symbol: str
    unrealizedFundingPnl: str
    unrealizedTradingPnl: str

@dataclass
class AccountInfo:
    assets: List[Asset]
    balance: str
    maximalWithdraw: str
    numFreeTransfersRemaining: int
    positions: List[Position]
    totalOrderNotional: str
    totalPositionNotional: str
    totalUnrealizedFundingPnl: str
    totalUnrealizedPnl: str
    totalUnrealizedTradingPnl: str
    tradeMakerFeeRate: str
    tradeTakerFeeRate: str

@dataclass
class AccountTrade:
    askAccountId: int
    askOrderId: int
    bidAccountId: int
    bidOrderId: int
    fee: str
    id: int
    orderType: str
    price: str
    quantity: str
    realizedPnl: str
    side: str
    symbol: str
    timestamp: int

@dataclass
class AccountTradesResponse:
    trades: List[AccountTrade]

@dataclass
class Settlement:
    direction: str
    indexPrice: str
    quantity: str
    settledAmount: str
    symbol: str
    timestamp: int

@dataclass
class SettlementsResponse:
    settlements: List[Settlement]




@dataclass
class PendingOrdersResponse:
    orders: List[Order]

@dataclass
class CapitalBalance:
    balance: str

@dataclass
class Transaction:
    id: int
    assetId: int
    quantity: str
    status: str
    timestamp: int
    transactionType: str
    transactionHash: Optional[Union[str,str]]
    token: Optional[str]
    etaTsSec: Optional[int]
    blockNumber: Optional[int]
    chain: Optional[str]
    instantWithdrawalChain: Optional[str]
    instantWithdrawalToken: Optional[str]
    isInstantWithdrawal: Optional[bool]
    withdrawalAddress: Optional[str]
    receivingAccountId: Optional[int]
    receivingAddress: Optional[str]
    srcAccountId: Optional[int]
    srcAddress: Optional[str]

    def __init__(self, 
                    id: int,
                    assetId: int,
                    quantity: str,
                    status: str,
                    timestamp: int,
                    transactionType: str,
                    transactionHash: Optional[Union[str,str]] = None,
                    token: Optional[str] = None,
                    etaTsSec: Optional[int] = None,
                    blockNumber: Optional[int] = None,
                    chain: Optional[str] = None,
                    instantWithdrawalChain: Optional[str] = None,
                    instantWithdrawalToken: Optional[str] = None,
                    isInstantWithdrawal: Optional[bool] = None,
                    withdrawalAddress: Optional[str] = None,
                    receivingAddress: Optional[str] = None,
                    receivingAccountId: Optional[int] = None,
                    srcAccountId: Optional[int] = None,
                    srcAddress: Optional[str] = None):
        self.id = id
        self.assetId = assetId
        self.quantity = quantity
        self.status = status
        self.timestamp = timestamp
        self.transactionType = transactionType
        self.transactionHash = transactionHash
        self.token = token
        self.etaTsSec = etaTsSec
        self.blockNumber = blockNumber
        self.chain = chain
        self.instantWithdrawalChain = instantWithdrawalChain
        self.instantWithdrawalToken = instantWithdrawalToken
        self.isInstantWithdrawal = isInstantWithdrawal
        self.withdrawalAddress = withdrawalAddress
        self.receivingAddress = receivingAddress
        self.receivingAccountId = receivingAccountId
        self.srcAccountId = srcAccountId
        self.srcAddress = srcAddress
@dataclass
class CapitalHistory:
    transactions: List[Transaction]

@dataclass
class WithdrawRequest:
    accountId: int
    coin: str
    withdrawAddress: str
    network: str
    quantity: str
    maxFees: str
    signature: str

@dataclass
class WithdrawResponse:
    orderId: str

@dataclass
class TransferRequest:
    accountId: int
    coin: str
    fees: str
    nonce: int
    quantity: str
    dstPublicKey: str
    signature: str

@dataclass
class TransferResponse:
    status: str

@dataclass
class DepositInfo:
    depositAddressEvm: str

class WebSocketSubscriptionTopic(Enum):
    MARK_PRICE = "mark_price"
    SPOT_PRICE = "spot_price"
    FUNDING_RATE_ESTIMATION = "funding_rate_estimation"
    TRADES = "trades"
    KLINES = "klines"
    ORDERBOOK = "orderbook"
    ASK_BID_PRICE ="ask_bid_price"
    

@dataclass
class WebSocketSubscription:
    symbol: str
    topic: WebSocketSubscriptionTopic

@dataclass
class WebSocketMarketSubscriptionListResponse:
    subscriptions: List[WebSocketSubscription]

#AUTOGEN BELOW

@dataclass
class WebSocketResponse:
    id: Optional[int]
    result: Optional[Dict[str, Any]]
    status: Optional[int]
    subscriptions: Optional[List[WebSocketSubscription]]

@dataclass
class WebSocketEvent:
    account: str
    event: str
    data: Dict[str, Any]

@dataclass
class WebSocketOrderCancelParams:
    orderId: str
    accountId: str
    nonce: int

@dataclass
class WebSocketOrderModifyParams:
    orderId: str
    accountId: str
    symbol: str
    quantity: str
    price: str
    maxFeesPercent: str

@dataclass
class WebSocketOrderStatusParams:
    orderId: str
    accountId: str

@dataclass
class WebSocketOrdersStatusParams:
    accountId: str

@dataclass
class WebSocketOrdersCancelParams:
    accountId: str
    nonce: str
    contractId: Optional[int] = None

@dataclass
class WebSocketBatchOrder:
    action: str
    nonce: int
    symbol: str
    orderType: str
    side: str
    quantity: str
    price: str
    maxFeesPercent: str
    signature: str
    orderId: Optional[str] = None
    updatedQuantity: Optional[str] = None
    updatedPrice: Optional[str] = None

@dataclass
class WebSocketOrdersBatchParams:
    accountId: str
    orders: List[WebSocketBatchOrder]

@dataclass
class WebSocketStreamStartParams:
    accountId: str

@dataclass
class WebSocketStreamPingParams:
    accountId: str
    listenKey: str
    timestamp: int

@dataclass
class WebSocketStreamStopParams:
    accountId: str
    listenKey: str
    timestamp: int

@dataclass
class AccountSnapshot:
    account_id: int
    balance: str
    positions: List[Position]

@dataclass
class AccountStreamStartResult:
    accountSnapshot: AccountSnapshot
    listenKey: str



@dataclass
class OrderPlaceParams:
    symbol: str
    quantity: float
    side: Side
    orderType: OrderType
    price: Optional[str]
    trigger_price: Optional[float]
    twap_config: Optional[TWAPConfig]
    maxFeesPercent: float
    orderFlags: Optional[str] = None
    creation_deadline: Optional[int] = None

@dataclass
class OrderCancelParams:
    orderId: str
    accountId: str
    nonce: int

@dataclass
class OrderModifyParams:
    orderId: str
    accountId: int
    symbol: str
    quantity: str
    price: str
    side: Side
    maxFeesPercent: str
    nonce: Optional[int]

@dataclass
class OrderStatusParams:
    orderId: str
    accountId: str

@dataclass
class OrdersStatusParams:
    accountId: int

@dataclass
class OrdersCancelParams:
    accountId: str
    nonce: str
    contractId: Optional[int] = None

@dataclass
class BatchOrder:
    action: str
    nonce: int
    symbol: Optional[str] = None
    orderType: Optional[OrderType] = None
    side: Optional[Side] = None
    quantity: Optional[str] = None
    price: Optional[str] = None
    maxFeesPercent: Optional[str] = None
    orderId: Optional[str] = None
    updatedQuantity: Optional[str] = None
    updatedPrice: Optional[str] = None
    signature: Optional[str] = None

@dataclass
class OrdersBatchParams:
    accountId: str
    orders: List[BatchOrder]

@dataclass
class EnableCancelOnDisconnectParams:
    nonce: int

@dataclass
class OrderResponse:
    accountId: str
    availableQuantity: str
    orderId: str
    orderType: OrderType
    price: str
    side: Side
    status: OrderStatus
    symbol: str
    totalQuantity: str


@dataclass
class OrderPlaceResponseResult:
    orderId: str

@dataclass
class OrderPlaceResponse:
    id:int
    result: OrderPlaceResponseResult
    status: int

@dataclass
class OrdersStatusResponse:
    id: int
    result: List[Order]
    status: Optional[int]

@dataclass
class OrderStatusResponse:
    id: int
    result: Order
    status: Optional[int]



@dataclass
class CrossChainAsset:
    chain: str
    exchangeRateFromUSDT: str
    exchangeRateToUSDT: str
    instantWithdrawalLowerLimitInUSDT: str
    instantWithdrawalUpperLimitInUSDT: str
    token: str



@dataclass
class TradingTier:
    level: int
    lowerThreshold: str
    title: str
    upperThreshold: str



@dataclass
class MarketInfo:
    category: str
    markPrice: str
    price24hAgo: str
    priceLatest: str
    tags: List[str]

@dataclass 
class Market:
    contract: FutureContract
    info: MarketInfo

@dataclass
class InventoryResponse:
    crossChainAssets: List[CrossChainAsset]
    feeConfig: FeeConfig
    markets: List[Market]
    tradingTiers: List[TradingTier]


import asyncio
from typing import Dict, List, Optional, Union

from websockets import ClientConnection, HeadersLike
import websockets
from hibachi_xyz.types import ExchangeInfo, FutureContract, MaintenanceWindow
from datetime import datetime
from prettyprinter import cpprint
from dataclasses import asdict, dataclass

default_api_url = "https://api.hibachi.xyz"
default_data_api_url = "https://data-api.hibachi.xyz"

async def connect_with_retry(web_url: str, headers: Optional[HeadersLike] = None) -> ClientConnection:
    """Establish WebSocket connection with retry logic"""
    max_retries = 10
    retry_count = 0
    retry_delay = 1 
    
    while retry_count < max_retries:
        try:
            websocket = await websockets.connect(web_url, 
                additional_headers=headers 
            )
            return websocket
        except Exception as e:
            retry_count += 1
            if retry_count >= max_retries:
                raise Exception(f"Failed to connect after {max_retries} attempts: {str(e)}")
            
            print(f"Connection attempt {retry_count} failed: {str(e)}. Retrying in {retry_delay} seconds...")
            await asyncio.sleep(retry_delay)
            retry_delay *= 2  # Exponential backoff
    
    # This shouldn't be reached due to the exception in the loop
    return websocket

def print_data(response):
    try:
        cpprint(asdict(response))
    except:
        cpprint(response)

def get_withdrawal_fee_for_amount(exchange_info: ExchangeInfo, amount: float) -> float:
    """
    Calculate the instant withdrawal fee for a given amount.
    
    Args:
        exchange_info: The exchange information
        amount: Withdrawal amount
        
    Returns:
        float: Fee percentage for the withdrawal
    """
    fees = exchange_info.feeConfig.instantWithdrawalFees
    # Sort fees by threshold (highest first)
    sorted_fees = sorted(fees, key=lambda x: x[0], reverse=True)
    
    for threshold, fee in sorted_fees:
        if amount >= threshold:
            return fee
    
    # Default to highest fee if amount is below all thresholds
    return sorted_fees[-1][1]

def get_next_maintenance_window(exchange_info: ExchangeInfo) -> Optional[MaintenanceWindow]:
    """
    Get the next maintenance window if any exists.
    
    Args:
        exchange_info: The exchange information
        
    Returns:
        Optional[Dict]: Details about the next maintenance window or None if none exists
    """
    windows = exchange_info.maintenanceWindow
    if not windows:
        return None
    
    now = datetime.now().timestamp()
    future_windows = [w for w in windows if w.begin > now]
    
    if not future_windows:
        return None
    
    next_window = min(future_windows, key=lambda w: w.begin)

    return next_window



def format_maintenance_window(window_info: MaintenanceWindow) -> str:
    """
    Format maintenance window information into a user-friendly string.
    
    Args:
        window_info: Maintenance window information from get_next_maintenance_window
        
    Returns:
        str: Formatted string with maintenance window details
    """
    if window_info is None:
        return "No upcoming maintenance windows scheduled."
    
    # Calculate time until maintenance starts
    now = datetime.now()
    start_time = datetime.fromtimestamp(window_info.begin)
    time_until = start_time - now
    
    duration_hours_raw = float((window_info.end - window_info.begin) / 3600.0)

    # Calculate days, hours, minutes
    days = time_until.days
    hours, remainder = divmod(time_until.seconds, 3600)
    minutes, _ = divmod(remainder, 60)
    
    # Format the start time
    start_str = start_time.strftime("%d %B %Y at %H:%M")
    
    # Format the duration
    if duration_hours_raw < 1:
        duration_str = f"{int(duration_hours_raw * 60)} minutes"
    else:
        duration_str = f"{int(duration_hours_raw)} hour{'s' if duration_hours_raw != 1 else ''}"
    
    # Combine all information
    return (
        f"The next maintenance window starts in {days}d{hours}h{minutes}m on {start_str} "
        f"for a duration of {duration_str}. "
        f"Reason: {window_info.note}."
    )


import hmac
import json
from dataclasses import asdict
from enum import Enum
from hashlib import sha256
from math import floor
from time import time, time_ns
from typing import Any, Dict, List, Optional, Tuple, TypeAlias, Union

import requests
from eth_keys import keys
from hibachi_xyz.helpers import (default_api_url, default_data_api_url,
                                 print_data)
from hibachi_xyz.types import (AccountInfo, AccountTrade,
                               AccountTradesResponse, Asset, BatchOrder,
                               BatchResponse, BatchResponseOrder,
                               CapitalBalance, CapitalHistory, CrossChainAsset,
                               DepositInfo, ExchangeInfo, FeeConfig,
                               FundingRateEstimation, FutureContract,
                               HibachiApiError, Interval, InventoryResponse,
                               Kline, KlinesResponse, MaintenanceWindow,
                               Market, MarketInfo, Nonce, OpenInterestResponse,
                               Order, OrderBook, OrderBookLevel, OrderId,
                               PendingOrdersResponse, Position, PriceResponse,
                               Settlement, SettlementsResponse, Side,
                               StatsResponse, TakerSide, Trade, TradesResponse,
                               TradingTier, Transaction, TransferRequest,
                               TransferResponse, TWAPConfig, WithdrawalLimit,
                               WithdrawRequest, WithdrawResponse)


class CreateOrder:
    action: str = "place"
    symbol: str
    side: Side
    quantity: float
    max_fees_percent: float
    price: Optional[float]
    trigger_price: Optional[float]
    twap_config: Optional[float]
    creation_deadline: Optional[float]

    def __init__(self, symbol: str, side: Side, quantity: float, max_fees_percent: float, price: Optional[float] = None, trigger_price: Optional[float] = None, twap_config: Optional[TWAPConfig] = None, creation_deadline: Optional[float] = None):

        if side == Side.BUY:
            side = Side.BID
        elif side == Side.SELL:
            side = Side.ASK

        self.symbol = symbol
        self.side = side
        self.quantity = quantity
        self.max_fees_percent = max_fees_percent
        self.price = price
        self.trigger_price = trigger_price
        self.twap_config = twap_config
        self.creation_deadline = creation_deadline

class UpdateOrder:
    action: str = "modify"
    order_id: int
    # needed for creating the signature
    symbol: str
    # needed for creating the signature
    side: Side
    quantity: float
    max_fees_percent: float
    price: Optional[float]
    trigger_price: Optional[float]

    def __init__(self, order_id: int, symbol: str, side: Side, quantity: float, max_fees_percent: float, price: Optional[float] = None, trigger_price: Optional[float] = None, creation_deadline: Optional[float] = None):

        if side == Side.BUY:
            side = Side.BID
        elif side == Side.SELL:
            side = Side.ASK

        self.order_id = order_id
        self.symbol = symbol
        self.side = side
        self.quantity = quantity
        self.max_fees_percent = max_fees_percent
        self.price = price
        self.trigger_price = trigger_price
        self.creation_deadline = creation_deadline

class CancelOrder:
    action: str = "cancel"
    order_id: Optional[int]
    nonce: Optional[int]

    def __init__(self, order_id: Optional[int] = None, nonce: Optional[int] = None):
        self.order_id = order_id
        self.nonce = nonce


def price_to_bytes(price: float, contract: FutureContract) -> bytes:
    return int(price * pow(2, 32) * pow(10, contract.settlementDecimals - contract.underlyingDecimals)).to_bytes(8, 'big')

def _get_http_error(response: requests.Response) -> Optional[HibachiApiError]:
    """ Check if the response is an error and return an exception if it is
        The builtin response.raise_for_status() does not show the server's response
    """

    if response.status_code > 299:
        return HibachiApiError(response.status_code, response.text)
    return None

class HibachiApiClient:
    """ 
    Example usage:
    ```python
    from hibachi_xyz import HibachiApiClient
    from dotenv import load_dotenv
    load_dotenv()

    hibachi = HibachiApiClient(
        api_key = os.environ.get('HIBACHI_API_KEY', "your-api-key"),
        account_id = os.environ.get('HIBACHI_ACCOUNT_ID', "your-account-id"),
        private_key = os.environ.get('HIBACHI_PRIVATE_KEY', "your-private"),
    )

    account_info = hibachi.get_account_info()
    print(f"Account Balance: {account_info.balance}")
    print(f"total Position Notional: {account_info.totalPositionNotional}")

    exchange_info = api.get_exchange_info()
    print(exchange_info)
    ```
    
    Args:
        api_url: The base URL of the API
        data_api_url: The base URL of the data API
        account_id: The account ID
        api_key: The API key
        private_key: The private key for the account
    
    """
    api_url: str
    data_api_url: str
    account_id: Optional[int] = None
    api_key: Optional[str] = None

    _private_key: Optional[keys.PrivateKey] = None # ECDSA for wallet account
    _private_key_hmac: Optional[str] = None # HMAC for web account

    future_contracts: Optional[Dict[str, FutureContract]] = None

    def __init__(self, api_url: str = default_api_url, data_api_url: str = default_data_api_url, 
            account_id: Optional[int] = None, 
            api_key: Optional[str] = None, 
            private_key: Optional[str] = None
            ):
        self.api_url = api_url
        self.data_api_url = data_api_url
        self.account_id = int(account_id) if isinstance(account_id, str) and account_id.isdigit() else account_id
        self.api_key = api_key
        if private_key is not None:
            self.set_private_key(private_key)

    def set_account_id(self, account_id: int):
        self.account_id = account_id

    def set_api_key(self, api_key: str):
        self.api_key = api_key

    def set_private_key(self, private_key: str):
        if private_key.startswith('0x'):
            private_key = private_key[2:]
            private_key_bytes = bytes.fromhex(private_key)
            self._private_key = keys.PrivateKey(private_key_bytes)

        if private_key.startswith('0x') == False:
            self._private_key_hmac = private_key
        
            

    """ Market API endpoints, can be called without having an account """

    def get_exchange_info(self) -> ExchangeInfo:
        """
        Return exchange metadata, currently it will return all futureContracts.

        Also returns a list of exchange maintenance windows in the "maintenanceWindow" field. For each window, the fields "begin" and "end" denote the beginning and end of the window, in seconds since the UNIX epoch. The field "note" contains a note.

        The field "maintenanceStatus" can have the values "NORMAL", "UNSCHEDULED_MAINTENANCE", "SCHEDULED_MAINTENANCE". If the exchange is currently under scheduled maintenance, the field "currentMaintenanceWindow" displays information on the current maintenance window.

        Endpoint: `GET /market/exchange-info`

        ```python
        exchange_info = client.get_exchange_info()
        print(exchange_info)
        ```
        Return type:
        ```python
        ExchangeInfo {
            feeConfig: FeeConfig
            futureContracts: List[FutureContract]
            instantWithdrawalLimit: WithdrawalLimit
            maintenanceWindow: List[MaintenanceWindow]
            # can be NORMAL, MAINTENANCE
            status: str
        }
        ```

        """
        exchange_info = self.__send_simple_request("/market/exchange-info")

        self.future_contracts = {}
        for contract in exchange_info["futureContracts"]:
            self.future_contracts[contract["symbol"]] = FutureContract(**contract)


        fee_config = FeeConfig(**exchange_info["feeConfig"])
    
        # Parse future contracts
        future_contracts = [FutureContract(**contract) for contract in exchange_info["futureContracts"]]
        
        # Parse withdrawal limit
        withdrawal_limit = WithdrawalLimit(**exchange_info["instantWithdrawalLimit"])
        
        # Parse maintenance windows
        maintenance_windows = [MaintenanceWindow(**window) for window in exchange_info["maintenanceWindow"]]
        
        # Create exchange info object
        return ExchangeInfo(
            feeConfig=fee_config,
            futureContracts=future_contracts,
            instantWithdrawalLimit=withdrawal_limit,
            maintenanceWindow=maintenance_windows,
            status=exchange_info["status"]
        )

        # exchange_info_types = ExchangeInfo(**exchange_info)

        # return exchange_info_types

    def get_inventory(self) -> InventoryResponse:
        """
        Similar to /market/exchange-info, in addition to the contract metadata we will return their latest price info.       

        Return type:
        ```python
        InventoryResponse {
            crossChainAssets: {
                chain: str
                exchangeRateFromUSDT: str
                exchangeRateToUSDT: str
                instantWithdrawalLowerLimitInUSDT: str
                instantWithdrawalUpperLimitInUSDT: str
                token: str
            }[]
            feeConfig: {
                depositFees: str
                instantWithdrawDstPublicKey: str
                instantWithdrawalFees: List[List[Union[int, float]]]
                tradeMakerFeeRate: str
                tradeTakerFeeRate: str
                transferFeeRate: str
                withdrawalFees: str
            }
            markets: {
                contract: {
                    displayName: str
                    id: int
                    maintenanceFactorForPositions: str
                    marketCloseTimestamp: Optional[str]
                    marketOpenTimestamp: Optional[str]
                    minNotional: str
                    minOrderSize: str
                    orderbookGranularities: List[str]
                    riskFactorForOrders: str
                    riskFactorForPositions: str
                    settlementDecimals: int
                    settlementSymbol: str
                    status: str
                    stepSize: str
                    symbol: str
                    tickSize: str
                    underlyingDecimals: int
                    underlyingSymbol: str
                }
                info: {
                    category: str
                    markPrice: str
                    price24hAgo: str
                    priceLatest: str
                    tags: List[str]
                }
            }[]
            tradingTiers: {
                level: int
                lowerThreshold: str
                title: str
                upperThreshold: str
            }[]
        }
        ```
        """
        market_inventory = self.__send_simple_request("/market/inventory")

        self.future_contracts = {}
        for market in market_inventory["markets"]:
            contract = FutureContract(**market["contract"])
            self.future_contracts[contract.symbol] = contract

        markets = [Market(contract=FutureContract(**m["contract"]), info=MarketInfo(**m["info"])) for m in market_inventory["markets"]]
                                                  

        output = InventoryResponse(
            crossChainAssets=[CrossChainAsset(**cca) for cca in market_inventory["crossChainAssets"]],
            feeConfig=FeeConfig(**market_inventory["feeConfig"]),
            markets=markets,
            tradingTiers=[TradingTier(**tt) for tt in market_inventory["tradingTiers"]]
        )

        return output

    def get_prices(self, symbol: str) -> PriceResponse:
        response = self.__send_simple_request(f"/market/data/prices?symbol={symbol}")
        response['fundingRateEstimation'] = FundingRateEstimation(**response['fundingRateEstimation'])
        return PriceResponse(**response)

    def get_stats(self, symbol: str) -> StatsResponse:
        return StatsResponse(**self.__send_simple_request(f"/market/data/stats?symbol={symbol}"))

    def get_trades(self, symbol: str) -> TradesResponse:
        response = self.__send_simple_request(f"/market/data/trades?symbol={symbol}")
        return TradesResponse(
            trades=[Trade(
                price=t["price"],
                quantity=t["quantity"],
                takerSide=TakerSide(t["takerSide"]),
                timestamp=t["timestamp"]
            ) for t in response["trades"]]
        )

    def get_klines(self, symbol: str, interval: Interval) -> KlinesResponse:
        response = self.__send_simple_request(f"/market/data/klines?symbol={symbol}&interval={interval.value}")
        return KlinesResponse(
            klines=[Kline(**kline) for kline in response["klines"]]
        )
    
    def get_open_interest(self, symbol: str) -> OpenInterestResponse:
        """Get open interest for a symbol

        Endpoint: `GET /market/data/open-interest`
        
        Args:
            symbol: The trading symbol (e.g. "BTC/USDT-P")
            
        Returns:
            OpenInterestResponse: The open interest data
        
        -----------------------------------------------------------------------
        """
        response = self.__send_simple_request(f"/market/data/open-interest?symbol={symbol}")
        return OpenInterestResponse(**response) 

    def get_orderbook(self, symbol: str, depth: int, granularity: float) -> OrderBook:
        """
        Get the orderbook price levels.
        It will return up to depth price levels on both side. The price level will be aggreated based on granularity.

        Endpoint: `GET /market/data/orderbook`

        Args:
            symbol: The trading symbol (e.g. "BTC/USDT-P")
            depth: The number of price levels to return on each side
            granularity: The price level granularity (e.g. 0.01)

        Return type:
        ```python
         OrderBook {
            ask: {
                price: str
                quantity: str
            }[]
            bid: {
                price: str
                quantity: str
            }[]
        }
        ```

        -----------------------------------------------------------------------
        """
        depth = int(depth)
        if depth < 1 or depth > 100:
            raise ValueError(
                "Depth must be a positive integer between 1 and 100, inclusive")

        self.__check_symbol(symbol)

        contract = self.future_contracts.get(symbol)
        granularities = contract.orderbookGranularities 
        if str(granularity) not in granularities:
            raise ValueError(
                f"Granularity for symbol {symbol} must be one of {granularities}")

        response = self.__send_simple_request(f"/market/data/orderbook?symbol={symbol}&depth={depth}&granularity={granularity}")
        
        ask_levels = [OrderBookLevel(price=level['price'], quantity=level['quantity']) for level in response['ask']['levels']]
        bid_levels = [OrderBookLevel(price=level['price'], quantity=level['quantity']) for level in response['bid']['levels']]
        
        return OrderBook(ask=ask_levels, bid=bid_levels)
    
    ### ===================================================== Account API =====================================================


    ### ------------------------------------------------ Account API - Capital ------------------------------------------------

    def get_capital_balance(self) -> CapitalBalance: 
        """
        Get the balance of your account.
        The returned balance is your net equity which includes unrealized PnL.

        Endpoint: `GET /capital/balance`

        ```python
        capital_balance = client.get_capital_balance()
        print(capital_balance.balance)
        ```        

        ```      
        CapitalBalance {
            balance: str
        }        
        ```
        -----------------------------------------------------------------------
        """
        self.__check_auth_data()
        response = self.__send_authorized_request('GET', f"/capital/balance?accountId={self.account_id}")        
        return CapitalBalance(**response)

    def get_capital_history(self) -> CapitalHistory:
        """
        Get the deposit and withdraw history of your account.
        It will return most recent up to 100 deposit and 100 withdraw.
        
        Endpoint: `GET /capital/history`

        ```python
        capital_history = client.get_capital_history()
        ```

        ```python
        Transaction {
            assetId: int
            blockNumber: int
            chain: Optional[str]
            etaTsSec: int
            id: int
            quantity: str
            status: str
            timestamp: int
            token: Optional[str]
            transactionHash: Union[str,str]
            transactionType: str
        }

        CapitalHistory {
            transactions: List[Transaction]
        }
        ```
        -----------------------------------------------------------------------
        """
        self.__check_auth_data()
        response = self.__send_authorized_request('GET', f"/capital/history?accountId={self.account_id}")

        return CapitalHistory(
            transactions=[Transaction(**tx) for tx in response['transactions']]
        )

    def withdraw(self, coin: str, withdraw_address: str, quantity: str, max_fees: str, network: str = "arbitrum") -> WithdrawResponse:
        """Submit a withdraw request.

        Endpoint: `POST /capital/withdraw`
        
        Args:
            coin: The coin to withdraw (e.g. "USDT")
            withdraw_address: The address to withdraw to
            quantity: The amount to withdraw should be no more than maximalWithdraw returned by /trade/account/info endpoint, otherwise it will be rejected.
            max_fees: Maximum fees allowed for the withdrawal
            network: The network to withdraw on (default "arbitrum")
            
        Returns:
            WithdrawResponse: The response containing the order ID
        
        -----------------------------------------------------------------------
        """
        self.__check_auth_data()
        
        # Create withdraw request payload
        request = WithdrawRequest(
            accountId=self.account_id,
            coin=coin,
            withdrawAddress=withdraw_address,
            network=network,
            quantity=quantity,
            maxFees=max_fees,
            signature=self.__sign_withdraw_payload(coin, withdraw_address, quantity, max_fees)
        )
        
        response = self.__send_authorized_request('POST', "/capital/withdraw", json=asdict(request))
        return WithdrawResponse(**response)

    def transfer(self, coin:str, quantity:str, dstPublicKey:str, max_fees:str):
        """
        Request fund transfer to another account.

        Endpoint: `POST /capital/transfer`

        Args:
            coin: The coin to transfer
            fees: The fees to transfer
            quantity: The quantity to transfer
        """

        nonce = time_ns() // 1_000

        request = TransferRequest(
            accountId=self.account_id,
            coin=coin,
            nonce=nonce,
            dstPublicKey=dstPublicKey.replace('0x', ''),
            fees=max_fees,
            quantity=quantity,
            signature=self.__sign_transfer_payload(nonce, coin, quantity, dstPublicKey, max_fees)
        )

        response = self.__send_authorized_request('POST', "/capital/transfer", json=asdict(request))

        return TransferResponse(**response)

    def get_deposit_info(self, public_key: str) -> DepositInfo:
        """Get deposit address information.

        Endpoint: `GET /capital/deposit-info`
        
        Args:
            public_key: The public key to get deposit info for
            
        Returns:
            DepositInfo: The deposit address information

        ```python
        DepositInfo { depositAddressEvm: str }
        ``` 
        -----------------------------------------------------------------------
        """
        response = self.__send_authorized_request('GET', f"/capital/deposit-info?accountId={self.account_id}&publicKey={public_key}")
        return DepositInfo(**response)

    def __sign_withdraw_payload(self, coin: str, withdraw_address: str, quantity: str, max_fees: str) -> str:
        """Sign a withdrawal request payload.
        
        Args:
            coin: The coin to withdraw
            withdraw_address: The withdrawal address
            quantity: The withdrawal amount
            max_fees: Maximum fees allowed
            
        Returns:
            str: The signature for the withdrawal request
        """
        # Get asset ID from exchange info
        if self.future_contracts is None:
            self.get_exchange_info()

           
        # Find asset ID for the coin
        asset_id = None
        for contract in self.future_contracts.values():
            if contract.settlementSymbol == coin:
                asset_id = contract.id
                break
                
        if asset_id is None:
            raise ValueError(f"Unknown coin: {coin}")

        # Create payload bytes
        asset_id_bytes = asset_id.to_bytes(4, 'big')
        quantity_bytes = int(float(quantity) * 1e6).to_bytes(8, 'big')  # Assuming 6 decimals for USDT
        max_fees_bytes = int(float(max_fees) * 1e6).to_bytes(8, 'big')  # Assuming 6 decimals for USDT
        address_bytes = bytes.fromhex(withdraw_address.replace('0x', ''))

        # Combine payload
        payload = asset_id_bytes + quantity_bytes + max_fees_bytes + address_bytes

        # Sign payload
        return self.__sign_payload(payload)
    
    def __sign_transfer_payload(self, nonce: int, coin: str, quantity: int, dst_account_public_key:str, max_fees_percent: str) -> str:
        # Get asset ID from exchange info
        if self.future_contracts is None:
            self.get_exchange_info()

           
        # Find asset ID for the coin
        asset_id = None
        for contract in self.future_contracts.values():
            if contract.settlementSymbol == coin:
                asset_id = contract.id
                break
                
        if asset_id is None:
            raise ValueError(f"Unknown coin: {coin}")

        # Create payload bytes
        nonce_bytes = nonce.to_bytes(8, 'big')
        asset_id_bytes = asset_id.to_bytes(4, 'big')
        quantity_bytes = int(float(quantity) * 1e6).to_bytes(8, 'big')  # Assuming 6 decimals for USDT
        max_fees_bytes = int(float(max_fees_percent)).to_bytes(8, 'big')
        address_bytes = bytes.fromhex(dst_account_public_key.replace('0x', ''))

        # Combine payload
        payload = nonce_bytes + asset_id_bytes + quantity_bytes + address_bytes + max_fees_bytes 

        # Sign payload
        return self.__sign_payload(payload)  
  

    ############################################################################
    ## Trade API endpoints, account_id and api_key must be set

    def get_account_info(self) -> AccountInfo:
        """
        Get account information/details

        Endpoint: `GET /trade/account/info`

        ```python
        account_info = client.get_account_info()
        print(account_info.balance)
        ```

        Return type:

        ```python
        AccountInfo {
            assets: {
                quantity: str
                symbol: str
            }[]
            balance: str
            maximalWithdraw: str
            numFreeTransfersRemaining: int
            positions: {
                direction: str
                entryNotional: str
                markPrice: str
                notionalValue: str
                openPrice: str
                quantity: str
                symbol: str
                unrealizedFundingPnl: str
                unrealizedTradingPnl: str
            }[]
            totalOrderNotional: str
            totalPositionNotional: str
            totalUnrealizedFundingPnl: str
            totalUnrealizedPnl: str
            totalUnrealizedTradingPnl: str
            tradeMakerFeeRate: str
            tradeTakerFeeRate: str
        }
        ```
        -----------------------------------------------------------------------
        """
        self.__check_auth_data()
        response = self.__send_authorized_request('GET', f"/trade/account/info?accountId={self.account_id}")
        
        assets = [Asset(**asset) for asset in response['assets']]
        positions = [Position(**position) for position in response['positions']]
        
        return AccountInfo(
            assets=assets,
            balance=response['balance'],
            maximalWithdraw=response['maximalWithdraw'],
            numFreeTransfersRemaining=response['numFreeTransfersRemaining'],
            positions=positions,
            totalOrderNotional=response['totalOrderNotional'],
            totalPositionNotional=response['totalPositionNotional'],
            totalUnrealizedFundingPnl=response['totalUnrealizedFundingPnl'],
            totalUnrealizedPnl=response['totalUnrealizedPnl'],
            totalUnrealizedTradingPnl=response['totalUnrealizedTradingPnl'],
            tradeMakerFeeRate=response['tradeMakerFeeRate'],
            tradeTakerFeeRate=response['tradeTakerFeeRate']
        )

    def get_account_trades(self) -> AccountTradesResponse:
        """
        Get the trades history of your account.
        It will return most recent up to 100 records.

        Endpoint: `GET /trade/account/trades`

        ```python
        account_trades = client.get_account_trades()
        ```

        Return type:

        ```python
        AccountTradesResponse {
            trades: {
                askAccountId: int
                askOrderId: int
                bidAccountId: int
                bidOrderId: int
                fee: str
                id: int
                orderType: str
                price: str
                quantity: str
                realizedPnl: str
                side: str
                symbol: str
                timestamp: int
            }[]
        }
        ```
        -----------------------------------------------------------------------
        """
        self.__check_auth_data()
        response = self.__send_authorized_request('GET', f"/trade/account/trades?accountId={self.account_id}")
        trades = [AccountTrade(**trade) for trade in response['trades']]
        return AccountTradesResponse(trades=trades)

    def get_settlements_history(self) -> SettlementsResponse:
        """
        You can obtain the history of settled trades.

        Endpoint: `GET /trade/account/settlements_history`

        ```python
        settlements = client.get_settlements_history()
        ```

        Return type:

        ```python
        SettlementsResponse {
            settlements: {
                direction: str
                indexPrice: str
                quantity: str
                settledAmount: str
                symbol: str
                timestamp: int
            }[]
        }
        ```
        -----------------------------------------------------------------------
        """
        self.__check_auth_data()
        response = self.__send_authorized_request('GET', f"/trade/account/settlements_history?accountId={self.account_id}")
        settlements = [Settlement(**settlement) for settlement in response['settlements']]
        return SettlementsResponse(settlements=settlements)

    def get_pending_orders(self) -> PendingOrdersResponse:
        """
        Get pending orders

        Endpoint: `GET /trade/orders`

        ```python
        pending_orders = client.get_pending_orders()
        ```

        Return type:
        ```python
        PendingOrdersResponse {
            orders: {
                accountId: int
                availableQuantity: str
                contractId: Optional[int]
                creationTime: Optional[int]
                finishTime: Optional[int]
                numOrdersRemaining: Optional[int]
                numOrdersTotal: Optional[int]
                orderId: str
                orderType: OrderType
                price: Optional[str]
                quantityMode: Optional[str]
                side: Side
                status: OrderStatus
                symbol: str
                totalQuantity: Optional[str]
                triggerPrice: Optional[str]
            }[]
        }    
        ```
        -----------------------------------------------------------------------
        """
        self.__check_auth_data()
        response = self.__send_authorized_request('GET', f"/trade/orders?accountId={self.account_id}")        
        orders = [Order(**order_data) for order_data in response]
        return PendingOrdersResponse(orders=orders)

    def get_order_details(self, order_id: Optional[int] = None, nonce: Optional[int] = None) -> Order:
        """
        Get order details

        Endpoint: `GET /trade/order`

        Either the order_id or the nonce can be used to query the order details

        ```python
        order_details = client.get_order_details(order_id=123)
        # or
        order_details = client.get_order_details(nonce=1234567)
        ```

        Return type:
        ```python
        Order {
            accountId: int
            availableQuantity: str
            contractId: Optional[int]
            creationTime: Optional[int]
            finishTime: Optional[int]
            numOrdersRemaining: Optional[int]
            numOrdersTotal: Optional[int]
            orderId: str
            orderType: OrderType
            price: Optional[str]
            quantityMode: Optional[str]
            side: Side
            status: OrderStatus
            symbol: str
            totalQuantity: Optional[str]
            triggerPrice: Optional[str]
        }
        ```        
        -----------------------------------------------------------------------
        """
        self.__check_order_selector(order_id, nonce)
        self.__check_auth_data()

        order_selector = f"orderId={order_id}" if order_id is not None else f"nonce={nonce}"
        response = self.__send_authorized_request('GET', f"/trade/order?accountId={self.account_id}&{order_selector}")
        
        response['numOrdersTotal'] = response.get('numOrdersTotal')
        response['numOrdersRemaining'] = response.get('numOrdersRemaining')
        response['totalQuantity'] = response.get('totalQuantity')
        response['quantityMode'] = response.get('quantityMode')
        response['price'] = response.get('price')
        response['triggerPrice'] = response.get('triggerPrice')       
        response['finishTime'] = response.get('finishTime')
        response['orderFlags'] = response.get('orderFlags')
        
        return Order(**response)

    # Order API endpoints require the private key to be set

    def place_market_order(self, symbol: str, quantity: float, side: Side, max_fees_percent: float, trigger_price: Optional[float] = None, twap_config: Optional[TWAPConfig] = None, creation_deadline: Optional[int] = None) -> tuple[Nonce, OrderId]:
        """
        Place a market order

        Endpoint: `POST /trade/order`

        ```python
        (nonce, order_id) = client.place_market_order("BTC/USDT-P", 0.0001, Side.BUY, max_fees_percent)
        (nonce, order_id) = client.place_market_order("BTC/USDT-P", 0.0001, Side.SELL, max_fees_percent)
        (nonce, order_id) = client.place_market_order("BTC/USDT-P", 0.0001, Side.BID, max_fees_percent, creation_deadline=2)
        (nonce, order_id) = client.place_market_order("BTC/USDT-P", 0.0001, Side.ASK, max_fees_percent, trigger_price=1_000_000)
        (nonce, order_id) = client.place_market_order("SOL/USDT-P", 1, Side.BID, max_fees_percent, twap_config=twap_config)
        (nonce, trigger_market_order_id) = client.place_market_order("BTC/USDT-P", 0.001, Side.ASK, max_fees_percent, trigger_price=90_100)
        ```
        """
        self.__check_auth_data()
        self.__check_symbol(symbol)

        if side == Side.BUY:
            side = Side.BID
        elif side == Side.SELL:
            side = Side.ASK

        if twap_config is not None and trigger_price is not None:
            raise ValueError("Can not set trigger price for TWAP order")

        nonce = time_ns() // 1_000
        request_data = self._create_order_request_data(nonce, symbol, quantity, side, max_fees_percent, trigger_price, None, creation_deadline, twap_config=twap_config)
        request_data["accountId"] = self.account_id
        response = self.__send_authorized_request('POST', f"/trade/order", json=request_data)
        order_id = int(response['orderId'])
        return (nonce, order_id)

    def place_limit_order(self, symbol: str, quantity: float, price: float, side: Side, max_fees_percent: float, trigger_price: Optional[float] = None, creation_deadline: Optional[int] = None) -> tuple[Nonce, OrderId]:
        """
        Place a limit order

        Endpoint: `POST /trade/order`

        ```python
        (nonce, order_id) = client.place_limit_order("BTC/USDT-P", 0.0001, 80_000, Side.BUY, max_fees_percent)
        (nonce, order_id) = client.place_limit_order("BTC/USDT-P", 0.0001, 80_000, Side.SELL, max_fees_percent)
        (nonce, order_id) = client.place_limit_order("BTC/USDT-P", 0.0001, 80_000, Side.BID, max_fees_percent, creation_deadline=2)
        (nonce, order_id) = client.place_limit_order("BTC/USDT-P", 0.0001, 1_001_000, Side.ASK, max_fees_percent, trigger_price=1_000_000)
        (nonce, limit_order_id) = client.place_limit_order("BTC/USDT-P", 0.001, 6_000, Side.BID, max_fees_percent)
        (nonce, trigger_limit_order_id) = client.place_limit_order("BTC/USDT-P", 0.001, 90_000, Side.ASK, max_fees_percent, trigger_price=90_100)
        ```
        """
        self.__check_auth_data()
        self.__check_symbol(symbol)

        if side == Side.BUY:
            side = Side.BID
        elif side == Side.SELL:
            side = Side.ASK

        nonce = time_ns() // 1_000
        request_data = self._create_order_request_data(nonce, symbol, quantity, side, max_fees_percent, trigger_price, price, creation_deadline)
        request_data["accountId"] = self.account_id
        response = self.__send_authorized_request('POST', f"/trade/order", json=request_data)
        order_id = int(response['orderId'])
        return (nonce, order_id)

    def update_order(self, order_id: int, max_fees_percent: float, quantity: Optional[float] = None, price: Optional[float] = None, trigger_price: Optional[float] = None, creation_deadline: Optional[int] = None) -> Dict[str, Any]:
        """
        Update an order

        Endpoint: `PUT /trade/order`

        ```python
        max_fees_percent = 0.0005   
        client.update_order(order_id, max_fees_percent, quantity=0.002)
        client.update_order(order_id, max_fees_percent, price=1_050_000)
        client.update_order(order_id, max_fees_percent, trigger_price=1_100_000)
        client.update_order(order_id, max_fees_percent, quantity=0.001, price=1_210_000, trigger_price=1_250_000)
        ```
        """
        self.__check_auth_data()
        order = self.get_order_details(order_id=order_id)
        symbol = order.symbol
        

        request_data_two = self._update_order_generate_sig(order, 
                                                          price=price, 
                                                          side=Side(order.side),
                                                          max_fees_percent=max_fees_percent,
                                                          trigger_price=trigger_price, 
                                                          quantity=quantity, 
                                                          creation_deadline=creation_deadline
                                                          )

        return self.__send_authorized_request('PUT', f"/trade/order", json=request_data_two)
    
    def _update_order_generate_sig(self, 
                                  order: Order, 
                                  side: Side,
                                  max_fees_percent: float,
                                  price: Optional[float] = None, 
                                  trigger_price: Optional[float] = None, 
                                  quantity: Optional[float] = None,
                                  creation_deadline: Optional[int] = None,
                                  nonce: Optional[Nonce] = None
                                  ) -> Dict[str, Any]:
        """ used to generate the signature for the update order request """
        symbol = order.symbol        
        self.__check_symbol(symbol)
        

        if order.orderType == 'MARKET' and price is not None:
            raise ValueError("Can not update price for a market order")

        if order.orderType == 'LIMIT' and price is None:
            price = float(order.price)

        if order.triggerPrice is None and trigger_price is not None:
            raise ValueError("Can not update trigger price for a non trigger order")

        if order.triggerPrice is not None and trigger_price is None:
            trigger_price = float(order.triggerPrice)

        if quantity is None:
            quantity = float(order.totalQuantity)

        side = Side(order.side)

        if side == Side.BUY:
            side = Side.BID
        elif side == Side.SELL:
            side = Side.ASK

        nonce = time_ns() // 1_000 if nonce is None else nonce
        request_data = self.__update_order_request_data(order.orderId, nonce, symbol, quantity, side, max_fees_percent, price=price, trigger_price=trigger_price, creation_deadline=creation_deadline)
        request_data["accountId"] = self.account_id
        return request_data


    def cancel_order(self, order_id: Optional[int] = None, nonce: Optional[int] = None) -> Dict[str, Any]:
        """
        Cancel an order

        Endpoint: `DELETE /trade/order`

        ```python
        client.cancel_order(order_id=123)
        client.cancel_order(nonce=1234567)
        ```
        """
        self.__check_order_selector(order_id, nonce)
        self.__check_auth_data()

        request_data = self._cancel_order_request_data(order_id, nonce, True)
        request_data["accountId"] = int(self.account_id)
        return self.__send_authorized_request('DELETE', f"/trade/order", json=request_data)

    def cancel_all_orders(self, contractId: Optional[int] = None) -> Dict[str, Any]:
        """
        Cancel all orders

        Endpoint: `DELETE /trade/orders`

        ```python
        client.cancel_all_orders()
        ```

        Note: currently there is a bug in the API where cancelling all orders is not working.
        This is a workaround to cancel all orders.
        """
        workaround = True

        if workaround:
            orders = self.get_pending_orders().orders
            for order in orders:
                self.cancel_order(order_id=int(order.orderId))
            return
        else:            
            self.__check_auth_data()
            nonce = time_ns() // 1_000
            request_data = self._cancel_order_request_data(None, nonce, False)
            request_data["accountId"] = int(self.account_id)
            return self.__send_authorized_request('DELETE', f"/trade/orders", json=request_data)

    def batch_orders(self, orders: list[CreateOrder | UpdateOrder | CancelOrder]) -> BatchResponse:
        """
        Creating, updating and cancelling orders can be done in a batch
        This requires knowing all details of the existing orders, there is no shortcut for update order details

        Endpoint: `POST /trade/orders`

        ```python
        response = client.batch_orders([
            # Simple market order
            CreateOrder("BTC/USDT-P", Side.SELL, 0.001, max_fees_percent),
            # Simple limit order
            CreateOrder("BTC/USDT-P", Side.SELL, 0.001, max_fees_percent, price=90_000),
            # Trigger market order
            CreateOrder("BTC/USDT-P", Side.SELL, 0.001, max_fees_percent, trigger_price=85_000),
            # Trigger limit order
            CreateOrder("BTC/USDT-P", Side.SELL, 0.001, max_fees_percent, price=84_750, trigger_price=85_000),
            # TWAP order
            CreateOrder("BTC/USDT-P", Side.SELL, 0.001, max_fees_percent, twap_config=TWAPConfig(5, TWAPQuantityMode.FIXED)),
            # Market order, only valid if placed within two seconds
            CreateOrder("BTC/USDT-P", Side.BUY, 0.001, max_fees_percent, creation_deadline=2),
            # Limit order, only valid if placed within one seconds
            CreateOrder("BTC/USDT-P", Side.BUY, 0.001, max_fees_percent, price=90_000, creation_deadline=1),
            # Trigger market order, only valid if placed within three seconds
            CreateOrder("BTC/USDT-P", Side.BUY, 0.001, max_fees_percent, trigger_price=85_000, creation_deadline=3),
            # Trigger limit order, only valid if placed within five seconds
            CreateOrder("BTC/USDT-P", Side.BUY, 0.001, max_fees_percent, price=75_250, trigger_price=75_000, creation_deadline=5),
            # TWAP order only valid if placed within two seconds
            CreateOrder("BTC/USDT-P", Side.SELL, 0.001, max_fees_percent, twap_config=TWAPConfig(5, TWAPQuantityMode.FIXED), creation_deadline=2),
            # Update limit order
            # Need to fill all relevant optional parameters
            UpdateOrder(limit_order_id, "BTC/USDT-P", Side.BUY, 0.001, max_fees_percent, price=60_000),
            # update trigger limit order
            # Need to fill all relevant optional parameters
            UpdateOrder(trigger_limit_order_id, "BTC/USDT-P", Side.ASK, 0.002, max_fees_percent, price=94_000, trigger_price=94_500),
            # update trigger market order
            # Need to fill all relevant optional parameters
            UpdateOrder(trigger_market_order_id, "BTC/USDT-P", Side.ASK, 0.001, max_fees_percent, trigger_price=93_000),
            # Cancel order
            CancelOrder(order_id=limit_order_id),
            CancelOrder(nonce=nonce),
        ])
        ```
        """
        self.__check_auth_data()

        nonce = time_ns() // 1_000
        orders_data = [self.__batch_order_request_data(nonce + i, order) for (i, order) in enumerate(orders)]
        request_data = {
            "accountId": int(self.account_id),
            "orders": orders_data
        }

        result = self.__send_authorized_request('POST', f"/trade/orders", json=request_data)
        allowed_keys = BatchResponseOrder.__dataclass_fields__.keys()
        orders = [
            BatchResponseOrder(**{k: v for k, v in order.items() if k in allowed_keys})
            for order in result["orders"]
        ]


        result['orders'] = orders
        return BatchResponse(**result)

    """ Private helpers """

    def __send_simple_request(self, path: str) -> Any:
        response = requests.get(f"{self.data_api_url}{path}")
        error = _get_http_error(response)
        if error is not None:
            raise error
        return response.json()

    def __check_auth_data(self):
        if self.account_id is None:
            raise RuntimeError("Account ID is not set")

        if self.api_key is None:
            raise RuntimeError("API key is not set")

    def __send_authorized_request(self, method: str, path: str, json: Optional[Any] = None) -> Any:
        headers = {
            "Authorization": self.api_key,
            "Content-Type": "application/json",
            "Accept": "application/json"
        }      

        response = requests.request(method, f"{self.api_url}{path}", headers=headers, json=json)
        error = _get_http_error(response)
        if error is not None:
            raise error

        return response.json()

    def __check_symbol(self, symbol: str):
        if self.future_contracts is None:
            self.get_exchange_info()

        if self.future_contracts.get(symbol) is None:
            raise ValueError(f"Unknown symbol: {symbol}")

    def __check_order_selector(self, order_id: Optional[int], nonce: Optional[int]):
        if order_id is None and nonce is None:
            raise ValueError("Either order_id or nonce must be provided")
        # if order_id is not None and nonce is not None:
        #     raise ValueError("Only one of order_id or nonce must be provided")

    def __sign_payload(self, payload: bytes) -> str:
        if self._private_key:
            # Hash the payload
            message_hash = sha256(payload).digest()

            # Sign the hash
            signed_message = self._private_key.sign_msg_hash(message_hash)

            # Extract signature components
            r = signed_message.r.to_bytes(32, 'big')
            s = signed_message.s.to_bytes(32, 'big')
            v = signed_message.v.to_bytes(1, 'big')

            # Combine to form the signature
            signature_hex = r.hex() + s.hex() + v.hex()

            return signature_hex
        
        if self._private_key_hmac:
            return hmac.new(self._private_key_hmac.encode(), payload, sha256).hexdigest()

        raise RuntimeError("Private key is not set")

    def __create_or_update_order_payload(self, contract: FutureContract, nonce: int, quantity: float, side: Side, max_fees_percent: float, price: Optional[float]) -> bytes:
        contract_id = contract.id

        nonce_bytes = nonce.to_bytes(8, 'big')
        contract_id_bytes = contract_id.to_bytes(4, 'big')
        quantity_bytes = int(quantity * pow(10, contract.underlyingDecimals)).to_bytes(8, 'big')
        price_bytes = b"" if price is None else price_to_bytes(price, contract)
        side_bytes = (0 if side.value == "ASK" else 1).to_bytes(4, 'big')
        max_fees_percent_bytes = int(max_fees_percent * pow(10, 8)).to_bytes(8, 'big')

        payload = nonce_bytes + contract_id_bytes + quantity_bytes + side_bytes + price_bytes + max_fees_percent_bytes

        return payload

    def _create_order_request_data(self, nonce: int, symbol: str, quantity: float, side: Side, max_fees_percent: float, trigger_price: Optional[float], price: Optional[float], creation_deadline: Optional[int], twap_config: Optional[TWAPConfig] = None) -> Dict[str, Any]:
        self.__check_auth_data()
        self.__check_symbol(symbol)
        contract = self.future_contracts.get(symbol)
        payload = self.__create_or_update_order_payload(contract, nonce, quantity, side, max_fees_percent, price)
        signature = self.__sign_payload(payload)

        if side == Side.BUY:
            side = Side.BID
        elif side == Side.SELL:
            side = Side.ASK

        request = {
            "nonce": nonce,
            "symbol": symbol,
            "quantity": str(quantity),
            "orderType": "MARKET",
            "side": side.value,
            "maxFeesPercent": str(max_fees_percent),
            "signature": signature,
        }
        if price is not None:
            request["orderType"] = "LIMIT"
            request["price"] = str(price)
        if trigger_price is not None:
            request["triggerPrice"] = str(trigger_price)
        if twap_config is not None:
            request = request | twap_config.to_dict()
        if creation_deadline is not None:
            deadline = floor(time()) + creation_deadline
            request["creationDeadline"] = deadline

   

        return request

    def __update_order_request_data(self, order_id: int, nonce: int, symbol: str, quantity: float, side: Side, max_fees_percent: float, price: Optional[float], trigger_price: Optional[float], creation_deadline: Optional[int]) -> Dict[str, Any]:
        contract = self.future_contracts.get(symbol)
        payload = self.__create_or_update_order_payload(contract, nonce, quantity, side, max_fees_percent, price)
        signature = self.__sign_payload(payload)
        request = {
            "nonce": nonce,
            "updatedQuantity": str(quantity),
            "quantity": str(quantity),
            "maxFeesPercent": str(max_fees_percent),
            "signature": signature,
        }
        if price is not None:
            request["updatedPrice"] = str(price)
            request["price"] = str(price)
        if order_id is not None:
            request["orderId"] = str(order_id)
        if trigger_price is not None:
            request["updatedTriggerPrice"] = str(trigger_price)
            request["trigger_price"] = str(trigger_price)
        if creation_deadline is not None:
            deadline = floor(time()) + creation_deadline
            request["creationDeadline"] = deadline
        return request

    def __cancel_order_payload(self, order_id: Optional[int], nonce: Optional[int]) -> bytes:
        if order_id is not None:
            return order_id.to_bytes(8, 'big')
        return nonce.to_bytes(8, 'big')

    def _cancel_order_request_data(self, order_id: Optional[int], nonce: Optional[int], nonce_as_str: bool) -> Dict[str, Any]:
        payload = self.__cancel_order_payload(order_id, nonce)
        signature = self.__sign_payload(payload)
        request = {
            "signature": signature
        }
        if order_id is not None:
            request["orderId"] = str(order_id)
        elif nonce_as_str:
            request["nonce"] = str(nonce)
        else:
            request["nonce"] = nonce
        return request

    def __batch_order_request_data(self, nonce: int, o: CreateOrder | UpdateOrder | CancelOrder) -> Dict[str, Any]:
        if type(o) is CreateOrder:
            payload = self._create_order_request_data(nonce, o.symbol, o.quantity, o.side, o.max_fees_percent, o.trigger_price, o.price, o.creation_deadline, twap_config=o.twap_config)
        elif type(o) is UpdateOrder:
            payload = self.__update_order_request_data(o.order_id, nonce, o.symbol, o.quantity, o.side, o.max_fees_percent, o.price, o.trigger_price, o.creation_deadline)
        else:
            payload = self._cancel_order_request_data(o.order_id, o.nonce, True)
        payload["action"] = o.action
        return payload



import asyncio

from hibachi_xyz import HibachiWSAccountClient, print_data
from hibachi_xyz.env_setup import setup_environment


async def example_ws_account_test(max_messages=3):
    api_endpoint, _, api_key, account_id, *_ = setup_environment()
    ws_url = api_endpoint.replace("https://", "wss://")
    
    client = HibachiWSAccountClient(
        api_endpoint=ws_url,
        api_key=api_key,
        account_id=account_id
    )
    
    await client.connect()
    await client.stream_start()
    
    received = []
    try:
        while len(received) < max_messages:
            message = await client.listen()
            if message:
                received.append(message)
    finally:
        await client.disconnect()

    return received



import asyncio
import contextlib
import json
from dataclasses import asdict
from typing import Callable, Dict, List, Optional

import websockets

from .helpers import connect_with_retry, default_data_api_url
from .types import WebSocketSubscription


class HibachiWSMarketClient:
    def __init__(self, api_endpoint: str = default_data_api_url):
        self.api_endpoint = api_endpoint.replace("https://", "wss://") + "/ws/market"
        self.websocket: Optional[websockets.WebSocketClientProtocol] = None
        self._event_handlers: Dict[str, List[Callable[[dict], None]]] = {}
        self._receive_task: Optional[asyncio.Task] = None

    async def connect(self):
        self.websocket = await connect_with_retry(self.api_endpoint)
        self._receive_task = asyncio.create_task(self._receive_loop())
        return self

    async def subscribe(self, subscriptions: List[WebSocketSubscription]):
        message = {
            "method": "subscribe",
            "parameters": {
                "subscriptions": [
                    {
                        **asdict(sub),
                        "topic": sub.topic.value
                    }
                    for sub in subscriptions
                ]
            }
        }
        await self.websocket.send(json.dumps(message))

    async def unsubscribe(self, subscriptions: List[WebSocketSubscription]):
        message = {
            "method": "unsubscribe",
            "parameters": {
                "subscriptions": [
                    {
                        **asdict(sub),
                        "topic": sub.topic.value
                    }
                    for sub in subscriptions
                ]
            }
        }
        await self.websocket.send(json.dumps(message))

    def on(self, topic: str, handler: Callable[[dict], None]):
        """Register a callback for raw topic name (e.g., 'mark_price')."""
        if topic not in self._event_handlers:
            self._event_handlers[topic] = []
        self._event_handlers[topic].append(handler)

    async def _receive_loop(self):
        try:
            while True:
                raw = await self.websocket.recv()
                msg = json.loads(raw)
                topic = msg.get("topic")
                if topic and topic in self._event_handlers:
                    for handler in self._event_handlers[topic]:
                        await handler(msg)
        except asyncio.CancelledError:
            pass
        except websockets.ConnectionClosed:
            print("[MarketClient] WebSocket closed.")
        except Exception as e:
            print(f"[MarketClient] Receive loop error: {e}")
            
    async def disconnect(self):
        if self._receive_task:
            self._receive_task.cancel()
            try:
                await self._receive_task
            except asyncio.CancelledError:
                pass
        if self.websocket:
            await self.websocket.close()
            self.websocket = None



from pathlib import Path
from importlib.metadata import version
import toml


from hibachi_xyz.api import *
from hibachi_xyz.types import *
from hibachi_xyz.helpers import *
from hibachi_xyz.api_ws_market import HibachiWSMarketClient
from hibachi_xyz.api_ws_trade import HibachiWSTradeClient
from hibachi_xyz.api_ws_account import HibachiWSAccountClient

def get_version() -> str:
    """Return the version of the hibachi library package."""
    pyproject_path = Path(__file__).parents[1] / "pyproject.toml"
    if pyproject_path.exists():
        with open(pyproject_path) as f:
            pyproject = toml.loads(f.read())

        return pyproject["project"]["version"]

    return version("hibachi_xyz")

__version__: str = get_version()
__all__ = [
    "HibachiApiClient",
    "HibachiApiError",
    "Interval",
    "Nonce",
    "OrderId",
    "Side",
    "TWAPConfig",
    "TWAPQuantityMode",
    "CreateOrder",
    "UpdateOrder",
    "CancelOrder",
    "__version__",
]



import asyncio
import json
import time
from typing import Callable, Dict, List, Optional

import websockets
from hibachi_xyz.helpers import connect_with_retry, default_api_url, print_data
from hibachi_xyz.types import (AccountSnapshot, AccountStreamStartResult,
                               Position)


class HibachiWSAccountClient:
    def __init__(self, api_key: str, account_id: str, api_endpoint: str = default_api_url):
        # self.api_endpoint = api_endpoint.replace("https://", "wss://") + "/ws/account"
        self.api_endpoint = api_endpoint.replace("https://", "wss://")
        self.websocket = None
        self.message_id = 0
        self.api_key = api_key
        self.account_id = int(account_id)
        self.listenKey: Optional[str] = None

    async def connect(self):
        self.websocket = await connect_with_retry(
            web_url=self.api_endpoint + f"/ws/account?accountId={self.account_id}",
            headers=[("Authorization", self.api_key)]
        )


    def _next_message_id(self) -> int:
        self.message_id += 1
        return self.message_id

    def _timestamp(self) -> int:
        return int(time.time())

    async def stream_start(self) -> AccountStreamStartResult:
        message = {
            "id": self._next_message_id(),
            "method": "stream.start",
            "params": {
                "accountId": self.account_id
            },
            "timestamp": self._timestamp()
        }

        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)

        result = AccountStreamStartResult(**response_data["result"])
        result.accountSnapshot = AccountSnapshot(**response_data["result"]["accountSnapshot"])
        result.accountSnapshot.positions = [Position(**pos) for pos in result.accountSnapshot.positions]
        self.listenKey = result.listenKey
        return result

    async def ping(self):
        if not self.listenKey:
            raise ValueError("Cannot send ping: listenKey not initialized.")

        message = {
            "id": self._next_message_id(),
            "method": "stream.ping",
            "params": {
                "accountId": self.account_id,
                "listenKey": self.listenKey
            },
            "timestamp": self._timestamp()
        }

        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        parsed = json.loads(response)
        if parsed.get("status") == 200:
            print("pong!")

    async def listen(self) -> Optional[dict]:
        try:
            response = await asyncio.wait_for(self.websocket.recv(), timeout=15)
            return json.loads(response)
        except asyncio.TimeoutError:
            await self.ping()
            return None
        except websockets.exceptions.ConnectionClosed as e:
            print(f"[listen] WebSocket closed: code={e.code}, reason={e.reason}")
            raise

    async def disconnect(self):
        if self.websocket:
            await self.websocket.close()
            self.websocket = None



import asyncio

from hibachi_xyz import (HibachiWSMarketClient, WebSocketSubscription,
                         WebSocketSubscriptionTopic, print_data)


async def example_ws_market_test():
    client = HibachiWSMarketClient()
    await client.connect()

    subscriptions = [
        WebSocketSubscription("BTC/USDT-P", WebSocketSubscriptionTopic.MARK_PRICE),
        WebSocketSubscription("BTC/USDT-P", WebSocketSubscriptionTopic.TRADES)
    ]
    await client.subscribe(subscriptions)

    messages = []

    try:
        for _ in range(5):
            msg = await client.websocket.recv()
            messages.append(msg)
    finally:
        await client.unsubscribe(subscriptions)
        await client.disconnect()

    return messages



import os
from pathlib import Path

from dotenv import load_dotenv


def setup_environment():
    # Load the .env file if it exists
    env_file_path = Path('.env')
    if env_file_path.exists():
        print(f"Loading environment variables from .env file")
        load_dotenv()  # This loads variables from .env file (if it exists)
    else:
        print(f".env file not found. Falling back to Bash Environment variables.")

    # Use a default environment if no environment is passed
    environment = os.getenv('ENVIRONMENT', 'production').lower()  # Default to 'production' if not passed

    # Print out the environment for debugging purposes
    print(f"Using {environment} environment")

    # Dynamically load environment variables based on the environment
    api_endpoint = os.environ.get(f'HIBACHI_API_ENDPOINT_{environment.upper()}', "https://api.hibachi.xyz")
    data_api_endpoint = os.environ.get(f'HIBACHI_DATA_API_ENDPOINT_{environment.upper()}', "https://data-api.hibachi.xyz")
    api_key = os.environ.get(f'HIBACHI_API_KEY_{environment.upper()}', "your-api-key")
    account_id = int(os.environ.get(f'HIBACHI_ACCOUNT_ID_{environment.upper()}', "your-account-id"))
    private_key = os.environ.get(f'HIBACHI_PRIVATE_KEY_{environment.upper()}', "your-private")
    public_key = os.environ.get(f'HIBACHI_PUBLIC_KEY_{environment.upper()}', "your-public")
    dst_public_key = os.environ.get(f'HIBACHI_TRANSFER_DST_ACCOUNT_PUBLIC_KEY_{environment.upper()}', "transfer-dst-account-public-key")

    # Return the environment variables for use in the tests
    return api_endpoint, data_api_endpoint, api_key, account_id, private_key, public_key, dst_public_key



import asyncio
import json
import os
import random
import time
from dataclasses import asdict
from typing import Any, Callable, Dict, List, Optional

import websockets
from eth_keys import keys
from hibachi_xyz.api import HibachiApiClient
from hibachi_xyz.helpers import (connect_with_retry, default_api_url,
                                 default_data_api_url, print_data)

from .types import (AccountInfo, AccountTradesResponse, CapitalBalance,
                    CapitalHistory, DepositInfo,
                    EnableCancelOnDisconnectParams, Nonce, Order,
                    OrderCancelParams, OrderModifyParams, OrderPlaceParams,
                    OrderPlaceResponse, OrderPlaceResponseResult,
                    OrderResponse, OrdersBatchParams, OrdersCancelParams,
                    OrdersStatusParams, OrdersStatusResponse,
                    OrderStatusParams, OrderStatusResponse,
                    PendingOrdersResponse, SettlementsResponse, Side,
                    WebSocketMarketSubscriptionListResponse, WebSocketResponse,
                    WebSocketSubscription, WebSocketSubscriptionTopic,
                    WithdrawRequest, WithdrawResponse)


class HibachiWSTradeClient:  
    """
    Trade Websocket Client is used to place, modify and cancel orders.
    
    ```python
    import asyncio
    import os
    from hibachi_xyz import HibachiWSTradeClient, print_data

    from dotenv import load_dotenv
    load_dotenv()

    account_id = int(os.environ.get('HIBACHI_ACCOUNT_ID', "your-account-id"))
    private_key = os.environ.get('HIBACHI_PRIVATE_KEY', "your-private")
    api_key = os.environ.get('HIBACHI_API_KEY', "your-api-key")
    public_key = os.environ.get('HIBACHI_PUBLIC_KEY', "your-public")

    async def main():
        client = HibachiWSTradeClient(
            api_key=api_key, 
            account_id=account_id, 
            account_public_key=public_key,
            private_key=private_key
        )

        await client.connect()    
        orders = await client.get_orders_status()
        first_order = orders.result[0]

        # single order
        order = await client.get_order_status(first_order.orderId)
        print_data(order)

        # client.api.set_private_key(private_key)
        modify_result = await client.modify_order(
            order=order.result,
            quantity=float("0.002"), 
            price=str(float("93500.0")),
            side=order.result.side,
            maxFeesPercent=float("0.00045"),
        )

        print_data(modify_result)

    asyncio.run(main())
    ```

    """
    def __init__(self,                  
                 api_key: str,                  
                 account_id: int, 
                 account_public_key: str,
                 api_url: str = default_api_url, 
                 data_api_url: str = default_data_api_url,
                 private_key: Optional[str] = None,                 
                ):
        self.api_endpoint = api_url
        self.api_endpoint = self.api_endpoint.replace("https://", "wss://") + "/ws/trade"
        self.websocket = None

        # random id start
        self.message_id = random.randint(1, 1000000)
        self._event_handlers: Dict[str, List[Callable]] = {}
        self._response_handlers: Dict[int, Callable] = {}
        self.api_key = api_key
        self.account_id = int(account_id) if isinstance(account_id, str) else account_id
        self.account_public_key = account_public_key

        self.api = HibachiApiClient(api_url=api_url, 
                                    data_api_url=data_api_url, 
                                    account_id=account_id, 
                                    api_key=api_key,
                                    private_key=private_key
                                    )

    async def connect(self):
        """Establish WebSocket connection with retry logic"""
        self.websocket = await connect_with_retry(
            web_url=self.api_endpoint + f"?accountId={self.account_id}", 
            headers=[("Authorization", self.api_key)]
        )

        return self

    async def place_order(self, params: OrderPlaceParams) -> tuple[Nonce,int]:
        """Place a new order"""
        self.message_id += 1

        nonce = time.time_ns() // 1_000
        side = params.side
        if side == Side.BUY:
            side = Side.BID
        elif side == Side.SELL:
            side = Side.ASK

        prepare_packet = self.api._create_order_request_data(
            nonce=nonce, 
            symbol=params.symbol, 
            quantity=params.quantity, 
            side=side, 
            max_fees_percent=params.maxFeesPercent, 
            trigger_price=params.trigger_price, 
            price=params.price, 
            creation_deadline=params.creation_deadline, 
            twap_config=params.twap_config)

        prepare_packet["accountId"] = self.account_id

        message = {
            "id": self.message_id,
            "method": "order.place",
            "params": prepare_packet,
            "signature": prepare_packet.get("signature")
        }


        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)

        print("ws place_order -------------------------------------------")
        print_data(response_data)

        # response_data["result"] = OrderPlaceResponseResult(**response_data["result"])
        # nonce: Nonce = prepare_packet.get("nonce")
        return (nonce, int(response_data.get("result").get("orderId")))

    async def cancel_order(self, orderId: int, nonce:int) -> WebSocketResponse:
        """Cancel an existing order"""
        self.message_id += 1

        prepare_packet = self.api._cancel_order_request_data(orderId, nonce, False)

        print("prepare_packet -------------------------------------------")
        print_data(prepare_packet)

        message = {
            "id": self.message_id,
            "method": "order.cancel",
            "params": {
                "orderId": str(orderId),
                "accountId": str(self.account_id),
                # "nonce": str(nonce)
            },
            "signature": prepare_packet.get("signature")
        }
        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)

        print_data(response_data)

        # return WebSocketResponse(**response_data)
        return response_data

    async def modify_order(self,
            order: Order,
            quantity: float,
            price: str,
            side: websockets.Side,
            maxFeesPercent: float,
            nonce: Optional[Nonce] = None
        ) -> WebSocketResponse:
        """Modify an existing order"""
        self.message_id += 1
        
        prepare_packet = self.api._update_order_generate_sig(
            order, 
            side=side, 
            max_fees_percent=maxFeesPercent, 
            quantity=quantity, 
            price=float(price),
            trigger_price=float(order.triggerPrice) if isinstance(order.triggerPrice, str) else order.triggerPrice,
            nonce=nonce
            )
        
        signature = prepare_packet.get("signature")
        del prepare_packet["signature"]

        message = {
            "id": self.message_id,
            "method": "order.modify",
            "params": prepare_packet,
            "signature": signature
        }

        

        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)       

        if "error" in response_data and response_data["error"]:
            raise Exception(f"Error modifying order: {response_data["error"]["message"]}")
        
        return response_data
        # return WebSocketResponse(**response_data)

    async def get_order_status(self, orderId: int) -> OrderStatusResponse:
        """Get status of a specific order"""
        self.message_id += 1
        message = {
            "id": self.message_id,
            "method": "order.status",
            "params": {
                "orderId" : str(orderId),
                "accountId" : int(self.account_id)
            }
        }

        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)

        print_data(response_data)

        response_data["result"] = Order(**response_data["result"])
        return OrderStatusResponse(**response_data)

    async def get_orders_status(self) -> OrdersStatusResponse:
        """Get status of all orders"""
        self.message_id += 1

        message = {
            "id": self.message_id,
            "method": "orders.status",
            "params": {
                "accountId" : int(self.account_id)
            }
        }

        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)
        response_data["result"] = [Order(**order) for order in response_data["result"]]
        return OrdersStatusResponse(**response_data)
        

    async def cancel_all_orders(self) -> bool:
        """Cancel all orders"""
        self.message_id += 1

        nonce = time.time_ns() // 1_000

        signed_packet = self.api._cancel_order_request_data(None, nonce, False)

        message = {
            "id": self.message_id,
            "method": "orders.cancel",
            "params": {
                "accountId": self.account_id,
                "nonce": nonce,
                # "contractId": 2 # TODO: get contract id
            }, 
            "signature": signed_packet.get("signature")
        }
        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)

        print_data(response_data)

        if response_data.get("id") == self.message_id:
            return response_data.get("status") == 200
        else:
            return False

    async def batch_orders(self, params: OrdersBatchParams) -> WebSocketResponse:
        """Execute multiple order operations in a single request"""
        self.message_id += 1
        message = {
            "id": self.message_id,
            "method": "orders.batch",
            "params": asdict(params)
        }
        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)
        return WebSocketResponse(**response_data)

    async def enable_cancel_on_disconnect(self, params: EnableCancelOnDisconnectParams) -> WebSocketResponse:
        """Enable automatic order cancellation on WebSocket disconnect"""
        self.message_id += 1
        message = {
            "id": self.message_id,
            "method": "orders.enableCancelOnDisconnect",
            "params": asdict(params)
        }
        await self.websocket.send(json.dumps(message))
        response = await self.websocket.recv()
        response_data = json.loads(response)
        return WebSocketResponse(**response_data)

    async def disconnect(self):
        """Close the WebSocket connection"""
        if self.websocket:
            await self.websocket.close()
            self.websocket = None

